﻿using System.Collections;

using UnityEngine;
using UnityEngine.UI;

/*
 * Class: BlackFading
 * 
 * A class that provides functionality to fade the screen to black
 */
[RequireComponent(typeof(CanvasRenderer))]
public class BlackFading : MonoBehaviour {
    public float fadingTime;

    // the black image
    private RawImage rawImage;

    // the component that provides alpha channel manipulation
    private CanvasRenderer canvasRenderer;

    // is the screen already black when initialized?
    public bool isBlackOnBeginning;

    // is the screen currently black?
    private bool isBlack = false;
    public static bool IsBlack => instance.isBlack;

    // is the screen currently fading to or from black?
    private bool isFading = false;
    public static bool IsFading => instance.isFading;

    // is the fading-system ready to use?
    public static bool IsReady => instance != null;

    // the singleton instance
    private static BlackFading instance;

    public void Awake() {
        if (instance == null) {
            instance = this;
            // get necessary components
            rawImage = GetComponent<RawImage>();
            canvasRenderer = GetComponent<CanvasRenderer>();

            // set values according to initial state
            if (isBlackOnBeginning) {
                isBlack = true;
                canvasRenderer.SetAlpha(1f);
                rawImage.raycastTarget = true;
            }
            else {
                isBlack = false;
                canvasRenderer.SetAlpha(0f);
                rawImage.raycastTarget = false;
            }
        }
    }

    /*
     * Start fading in the black screen
     */
    public static void FadeIn() {
        if (instance != null) {
            instance.isFading = true;
            instance.StopAllCoroutines();
            instance.StartCoroutine(instance.FadeInCoroutine());
        }
    }

    /*
     * The coroutine that handles the fading to black
     */
    public IEnumerator FadeInCoroutine() {
        isFading = true;
        rawImage.raycastTarget = true;
        float fadingFactor = 1f / fadingTime;
        while (canvasRenderer.GetAlpha() < 1f) {
            canvasRenderer.SetAlpha(Mathf.Min(1f, canvasRenderer.GetAlpha() + Time.unscaledDeltaTime * fadingFactor));
            if (canvasRenderer.GetAlpha() != 1f) {
                yield return null;
            }
        }
        isFading = false;
        isBlack = true;
    }

    /*
     * Start fading out the black screen
     */
    public static void FadeOut() {
        if (instance != null) {
            instance.isFading = true;
            instance.StopAllCoroutines();
            instance.StartCoroutine(instance.FadeOutCoroutine());
        }
    }

    /*
     * The coroutine that handles the fading from black
     */
    public IEnumerator FadeOutCoroutine() {
        isFading = true;
        float fadingFactor = 1f / fadingTime;
        while (canvasRenderer.GetAlpha() > 0f) {
            canvasRenderer.SetAlpha(Mathf.Max(0f, canvasRenderer.GetAlpha() - Time.unscaledDeltaTime * fadingFactor));
            if (canvasRenderer.GetAlpha() != 0f) {
                yield return null;
            }
        }
        rawImage.raycastTarget = false;
        isFading = false;
        isBlack = false;
    }
}
