﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class NavMesh : MonoBehaviour
{
    public enum Direction {
        UP, RIGHT, DOWN, LEFT
    }

    [System.Serializable]
    public class Position : System.IEquatable<Position> {
        public int row;
        public int col;
        public Vector2 Coordinates => NavMesh.GetPositionCoordinates(this);
        public Position left => col - 1 >= 0 ? new Position(row, col - 1) : null;
        public Position right => col + 1 < GameManager.Instance.navMeshSize.width ? new Position(row, col + 1) : null;
        public Position up => row + 1 < GameManager.Instance.navMeshSize.height ? new Position(row + 1, col) : null;
        public Position down => row - 1 >= 0 ? new Position(row - 1, col) : null;
        public bool Equals(Position other)
        {
            if (other == null)
            {
                return false;
            }
            return row == other.row && col == other.col;
        }
        public Position GetNeighbor(Direction dir)
        {
            switch (dir)
            {
                case Direction.DOWN:
                    return down;
                case Direction.LEFT:
                    return left;
                case Direction.RIGHT:
                    return right;
                case Direction.UP:
                    return up;
                default:
                    return null;
            }
        }
        public Position()
        {
            this.row = 0;
            this.col = 0;
        }
        public Position(int row, int col) {
            this.row = row;
            this.col = col;
        }
    }

    public NavMeshSize navMeshSize;
    public int Width => (navMeshSize == null ? 1 : navMeshSize.width);
    public int Height => (navMeshSize == null ? 1 : navMeshSize.height);

    public Position Entrance => new Position(0, Width / 2);
    public Position PlayerStart => new Position(12, Width / 2);

    private bool[,] navMeshBlocked;
    public bool[,] NavMeshBlocked => (bool[,])navMeshBlocked.Clone();

    public void Awake() {
        navMeshBlocked = new bool[Height, Width];
    }

    public static Vector2 GetPositionCoordinates(Position pos) {
        return new Vector2(-GameManager.Instance.navMeshSize.width / 2.0f + pos.col + 0.5f, -GameManager.Instance.navMeshSize.height / 2.0f + pos.row + 0.5f);
    }

    public void FreeNavMeshAt(Position pos)
    {
        if (navMeshBlocked != null)
            navMeshBlocked[pos.row, pos.col] = false;
    }

    public void BlockNavMeshAt(Position pos)
    {
        if (navMeshBlocked != null)
            navMeshBlocked[pos.row, pos.col] = true;
    }

    public Position GetNewPosition(Position pos, Direction dir)
    {
        int offset_horizontal = 0;
        int offset_vertical = 0;
        switch (dir)
        {
            case Direction.UP:
                offset_vertical = 1;
                break;
            case Direction.RIGHT:
                offset_horizontal = 1;
                break;
            case Direction.DOWN:
                offset_vertical = -1;
                break;
            case Direction.LEFT:
                offset_horizontal = -1;
                break;
        }
        return new Position(pos.row + offset_vertical, pos.col + offset_horizontal);
    }

    public bool GetWalkable(Position pos, Direction dir) {
        Position newPos = GetNewPosition(pos, dir);
        return newPos.row >= 0 && newPos.row < Height && newPos.col >= 0 && newPos.col < Width && !navMeshBlocked[newPos.row, newPos.col];
    }

    public void OnDrawGizmos()
    {
        Gizmos.color = Color.yellow;
        Gizmos.DrawSphere(Entrance.Coordinates, 0.4f);
        Gizmos.color = Color.blue;
        Gizmos.DrawSphere(PlayerStart.Coordinates, 0.4f);
        float left = -Width / 2.0f;
        float bottom = -Height / 2.0f;
        for (int i = 0; i <= Width; ++i) {
            Gizmos.DrawLine(new Vector3(left + i, bottom, 0f), new Vector3(left + i, bottom + Height, 0f));
        }
        for (int i = 0; i <= Height; ++i) {
            Gizmos.DrawLine(new Vector3(left, bottom + i, 0f), new Vector3(left + Width, bottom + i, 0f));
        }
        if (navMeshBlocked != null)
        {
            for (int i = 0; i < Width; ++i)
            {
                for (int j = 0; j < Height; ++j)
                {
                    Gizmos.color = navMeshBlocked[j, i] ? Color.red : Color.green;
                    Gizmos.DrawSphere(new Vector3(left + i + 0.5f, bottom + j + 0.5f, 0f), 0.3f);
                }
            }
        }
    }
}
