﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class Person : MonoBehaviour
{
    public SpriteRenderer personSprite;

    public NavMesh.Position currentPosition;
    public float unitsPerSecond;

    protected NavMesh.Direction lookingDirection;
    protected bool isMoving;
    public bool IsMoving => isMoving;
    protected bool isNavigating;

    public bool IsNavigating => isNavigating;
    protected Vector2 targetCoordinates;
    public NavMesh.Position targetPosition;

    public virtual void Turn(NavMesh.Direction dir)
    {
        lookingDirection = dir;
    }

    public virtual void MoveTo(NavMesh.Position newPos)
    {
        if (newPos.Equals(currentPosition.left))
        {
            Turn(NavMesh.Direction.LEFT);
        }
        else if (newPos.Equals(currentPosition.up))
        {
            Turn(NavMesh.Direction.UP);
        }
        else if (newPos.Equals(currentPosition.right))
        {
            Turn(NavMesh.Direction.RIGHT);
        }
        else
        {
            Turn(NavMesh.Direction.DOWN);
        }
        targetCoordinates = NavMesh.GetPositionCoordinates(newPos);
        isMoving = true;
        GameManager.Instance.navMesh.FreeNavMeshAt(currentPosition);
        GameManager.Instance.navMesh.BlockNavMeshAt(newPos);
        currentPosition = newPos;
        int layer = 1000 - 10 * currentPosition.row;
        personSprite.sortingOrder = layer;
    }

    public void NavigateTo(NavMesh.Position targetPos)
    {
        if (!targetPos.Equals(currentPosition))
        {
            isNavigating = true;
            targetPosition = targetPos;
            NavigationStep();
        }
    }

    protected void NavigationStep()
    {
        StartCoroutine(KeepTrying());
    }

    public IEnumerator KeepTrying()
    {
        NavMesh.Position nextPosition = GetNextPosition();
        while (nextPosition == null)
        {
            yield return null;
            nextPosition = GetNextPosition();
        }
        MoveTo(nextPosition);
    }

    private int GetMinIndex(List<float> floats)
    {
        int minIdx = -1;
        float minVal = float.MaxValue;
        for (int i = 0; i < floats.Count; ++i)
        {
            if (minIdx == -1 || floats[i] < minVal)
            {
                minIdx = i;
                minVal = floats[i];
            }
        }
        return minIdx;
    }

    public NavMesh.Position GetNextPosition()
    {
        bool[,] navMeshBlocks = GameManager.Instance.navMesh.NavMeshBlocked;
        NavMesh.Position[,] predecessors = new NavMesh.Position[navMeshBlocks.GetLength(0), navMeshBlocks.GetLength(1)];
        float[,] weights = new float[navMeshBlocks.GetLength(0), navMeshBlocks.GetLength(1)];
        for (int i = 0; i < weights.GetLength(0); ++i)
        {
            for (int j = 0; j < weights.GetLength(1); ++j)
            {
                weights[i, j] = 1000;
            }
        }
        List<NavMesh.Position> openList = new List<NavMesh.Position>();
        List<float> priorities = new List<float>();
        List<NavMesh.Position> closedList = new List<NavMesh.Position>();

        //A* starting here
        openList.Add(currentPosition);
        priorities.Add(0f);

        do
        {
            int minIdx = GetMinIndex(priorities);
            float priority = priorities[minIdx];
            NavMesh.Position current = openList[minIdx];
            priorities.RemoveAt(minIdx);
            openList.RemoveAt(minIdx);

            if (current.Equals(targetPosition))
            {
                while (!predecessors[current.row, current.col].Equals(currentPosition))
                {
                    current = predecessors[current.row, current.col];
                }
                return current;
            }

            closedList.Add(current);

            NavMesh.Position[] neighbors = new NavMesh.Position[] { current.left, current.up, current.right, current.down };
            NavMesh.Direction[] dirs = new NavMesh.Direction[] { NavMesh.Direction.LEFT, NavMesh.Direction.UP, NavMesh.Direction.RIGHT, NavMesh.Direction.DOWN };
            for (int i = 0; i < 4; ++i)
            {
                NavMesh.Position neighbor = neighbors[i];
                NavMesh.Direction dir = dirs[i];
                if (neighbor != null && GameManager.Instance.navMesh.GetWalkable(current, dir) && !closedList.Contains(neighbor))
                {
                    float weightComb = weights[current.row, current.col] + 1;
                    int neighborIdx = openList.IndexOf(neighbor);
                    if (neighborIdx == -1 || weightComb < weights[neighbor.row, neighbor.col])
                    {
                        predecessors[neighbor.row, neighbor.col] = current;
                        weights[neighbor.row, neighbor.col] = weightComb;
                        priority = weightComb + Mathf.Sqrt(Mathf.Pow(targetPosition.col - current.col, 2) + Mathf.Pow(targetPosition.row - current.col, 2));
                        if (neighborIdx == -1)
                        {
                            openList.Add(neighbor);
                            priorities.Add(priority);
                        }
                        else
                        {
                            priorities[neighborIdx] = priority;
                        }
                    }
                }
            }
        } while (openList.Count > 0);

        return null;
    }

    public virtual void ArrivalCallback()
    {

    }

    public virtual void NavigationEndCallback()
    {

    }
    
    protected void Update()
    {
        if (isMoving)
        {
            transform.position = Vector2.MoveTowards(transform.position, targetCoordinates, unitsPerSecond * Time.deltaTime);
            if (transform.position.Equals(targetCoordinates))
            {
                isMoving = false;
                ArrivalCallback();
                if (isNavigating)
                {
                    if (currentPosition.Equals(targetPosition))
                    {
                        isNavigating = false;
                        NavigationEndCallback();
                    }
                    else
                    {
                        NavigationStep();
                    }
                }
            }
        }
    }
}
